import logging
from typing import Optional

from github_webhooks.schemas import WebhookCommonPayload, WebhookHeaders


class PingPayload(WebhookCommonPayload):
    zen: str
    hook_id: int


async def handle_ping(payload: PingPayload, headers: WebhookHeaders) -> Optional[str]:
    logging.debug('ping event, headers: %s, payload: %s', headers, payload)
    return 'OK'



from typing import Optional

from pydantic import BaseModel


class User(BaseModel):
    id: int  # noqa: A003
    type: str  # noqa: A003
    login: str


class Organization(BaseModel):
    id: int  # noqa: A003
    type: str  # noqa: A003
    login: str


class Repository(BaseModel):
    id: int  # noqa: A003
    name: str
    full_name: str
    fork: bool

    owner: User
    organization: Optional[Organization]



import logging
from typing import Optional

from github_webhooks.schemas import WebhookHeaders


async def handle_default(event: str, payload: bytes, headers: WebhookHeaders) -> Optional[str]:
    logging.debug('Default handler for <%s> event, headers: %s, payload: %s', event, headers, payload)
    return 'OK'



from . import ping
from .registry import HandlersRegistry

__all__ = (
    'HandlersRegistry',
    'registry',
)


registry = HandlersRegistry()
registry.add_handler('ping', ping.PingPayload, ping.handle_ping)



import hmac
import logging
from typing import Optional

from fastapi import APIRouter, HTTPException, Request
from pydantic import parse_obj_as

from .schemas import WebhookHeaders

router = APIRouter()


@router.post('/hook')
async def webhook_handler(request: Request) -> str:
    headers = parse_obj_as(WebhookHeaders, request.headers)

    payload_body: bytes = await request.body()

    verify_signature(payload_body, headers, request.app.secret_token)

    result = await request.app.hooks.handle(headers.event, payload_body, headers)

    return result or 'OK'


def verify_signature(payload_body: bytes, headers: WebhookHeaders, secret_token: Optional[str]) -> None:

    if not secret_token:
        # validation is disabled
        return

    if not headers.signature_256:
        logging.error('Request is missing signature header')
        raise HTTPException(status_code=403, detail='signature verification failed')

    signature = hmac.new(secret_token.encode('utf-8'), payload_body, 'sha256').hexdigest()

    if not hmac.compare_digest(f'sha256={signature}', headers.signature_256):
        logging.error(
            'Signature verification failed "%s" != "%s"',
            f'sha256={signature}',
            headers.signature_256,
        )
        raise HTTPException(status_code=403, detail='signature verification failed')



from typing import Any, Callable, Union, cast

from github_webhooks.schemas import WebhookHeaders

from .default import handle_default
from .types import AnyHandlerWithHeaders, AnyHandlerWithoutHeaders, DefaultHandler, Handler, HandlerResult, PayloadT


class HandlersRegistry:
    _handlers: dict[str, tuple[PayloadT, Handler]]
    _default_handler: DefaultHandler

    def __init__(self) -> None:
        self._handlers = {}
        self._default_handler = handle_default

    def set_default_handler(self, handler: DefaultHandler) -> None:
        self._default_handler = handler

    def add_handler(self, event: str, payload_cls: PayloadT, handler: Handler) -> None:
        self._handlers[event] = (payload_cls, handler)

    def register(self, event: str, payload_cls: PayloadT) -> Callable[[Handler], Handler]:
        def deco(func: Handler) -> Handler:
            self.add_handler(event, payload_cls, func)
            return func

        return deco

    async def handle(self, event: str, payload: bytes, headers: WebhookHeaders) -> HandlerResult:
        if event not in self._handlers:
            return await self._call_with_headers(self._default_handler, event, payload, headers=headers)

        payload_cls, handler = self._handlers[event]

        payload_parsed = payload_cls.parse_raw(payload)
        return await self._call_with_headers(handler, payload_parsed, headers=headers)

    @staticmethod
    async def _call_with_headers(
        handler: Union[DefaultHandler, Handler],
        *args: Any,
        headers: WebhookHeaders,
    ) -> HandlerResult:
        h_varnames = handler.__code__.co_varnames  # type: ignore

        if 'headers' in h_varnames:
            handler = cast(AnyHandlerWithHeaders, handler)
            return await handler(*args, headers=headers)

        handler = cast(AnyHandlerWithoutHeaders, handler)
        return await handler(*args)



from typing import Optional

from fastapi import FastAPI

from . import handlers, web


class App(FastAPI):
    secret_token: Optional[str] = None

    hooks: handlers.HandlersRegistry


def create_app(secret_token: Optional[str] = None) -> App:
    app = App()

    app.hooks = handlers.registry
    app.secret_token = secret_token

    app.include_router(web.router)

    return app



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['github_webhooks', 'github_webhooks.handlers', 'github_webhooks.schemas']

package_data = \
{'': ['*']}

install_requires = \
['fastapi>=0.73.0,<0.74.0', 'pydantic>=1.9.0,<2.0.0']

setup_kwargs = {
    'name': 'github-webhooks-framework',
    'version': '0.1.7',
    'description': 'GitHub Webhooks Framework',
    'long_description': "# Python GitHub Webhooks Framework\n\nSimple and lightweight micro framework for quick integration with [GitHub\nwebhooks][1].  \nIt's based on [FastAPI][3] and [pydantic][4], nothing more!  \nAsync and mypy friendly. \n\n[![Run CI](https://github.com/karech/github-webhooks/actions/workflows/ci.yml/badge.svg?branch=main)](https://github.com/karech/github-webhooks/actions/workflows/ci.yml)  \n[![PyPI](https://img.shields.io/pypi/v/github-webhooks-framework.svg)][2]\n\n\n## Installation\nJust add `github-webhooks-framework` package.   \nExample: \n* `pip install github-webhooks-framework`\n* `poetry add github-webhooks-framework`\n\n\n## Example\nCreate file `example.py` and copy next code:\n```python\nimport uvicorn\nfrom pydantic import BaseModel\n\nfrom github_webhooks import create_app\nfrom github_webhooks.schemas import WebhookCommonPayload\n\n\n# WebhookCommonPayload is based on pydantic.BaseModel\nclass PullRequestPayload(WebhookCommonPayload):\n    class Pull(BaseModel):\n        title: str\n        url: str\n    \n    action: str\n    pull_request: Pull\n    \n\n# Initialize Webhook App\napp = create_app()\n\n\n# Register webhook handler:\n#   `pull_request` - name of an event to handle\n#   `PullRequestPayload` - webhook payload will be parsed into this model\n@app.hooks.register('pull_request', PullRequestPayload)\nasync def handler(payload: PullRequestPayload) -> None:\n    print(f'New pull request {payload.pull_request.title}')\n    print(f'  link: {payload.pull_request.url}')\n    print(f'  author: {payload.sender.login}')\n\n\nif __name__ == '__main__':\n    # start uvicorn server\n    uvicorn.run(app)\n```\n  \n \n### Let's have detailed overview. \n\nWe start by defining payload [Model][5] to parse incoming [Pull Request Body][6]. \n```python\nclass PullRequestPayload(WebhookCommonPayload):\n    class Pull(BaseModel):\n        title: str\n        url: str\n    \n    action: str\n    pull_request: Pull\n```\nIn this example we only want to get `action`, `pull_request.title` and `pull_request.url` from payload.  \nBy subclassing `WebhookCommonPayload` model will automatically get `sender`, `repository` and `organization` fields.\n\n\nNext - we are creating ASGI app (based on FastAPI app)\n```python\napp = create_app()\n```\nOptionally we can provide here `secret_token` [Github Webhook secret][7]\n```python\napp = create_app(secret_token='super-secret-token')\n```\n\nAnd time to define our handler\n```python\n@app.hooks.register('pull_request', PullRequestPayload)\nasync def handler(payload: PullRequestPayload) -> None:\n    print(f'New pull request {payload.pull_request.title}')\n    print(f'  link: {payload.pull_request.url}')\n    print(f'  author: {payload.sender.login}')\n```\n\nWe are using here `@app.hooks.register` deco, which accepts 2 arguments:\n* `event: str` - name of webhook event\n* `payload_cls: pydantic.BaseModel` - pydantic model class to parse request, subclassed from `pydantic.BaseModel` \nor `WebhookCommonPayload`.\n\nAnd our handler function must be any of this signatures:  \n```python\nasync def handler(payload: PullRequestPayload) -> None:\n    ...\n```\n```python\nasync def handler(payload: PullRequestPayload, headers: WebhookHeaders) -> Optional[str]:\n    # `headers` will be WebhookHeaders model with Github Webhook headers parsed.\n    ...\n```  \n\n\nAnd the last - let's launch it.  \nFor example with uvicorn  \n```shell\nuvicorn example:app\n```\nWebhook will be available on http://localhost:8000/hook\n\nThat's it!\nNow you have a webhook server, which can handle incoming Github Webhook requests.\n\n\n\n[1]: https://developer.github.com/webhooks/\n[2]: https://pypi.python.org/pypi/github-webhooks-framework\n[3]: https://fastapi.tiangolo.com/\n[4]: https://pydantic-docs.helpmanual.io/\n[5]: https://pydantic-docs.helpmanual.io/usage/models/\n[6]: https://docs.github.com/en/developers/webhooks-and-events/webhooks/webhook-events-and-payloads#pull_request\n[7]: https://docs.github.com/en/developers/webhooks-and-events/webhooks/securing-your-webhooks\n",
    'author': 'karech',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/karech/github-webhooks',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from typing import Optional

from pydantic import BaseModel, Field

from .objects import Organization, Repository, User


class WebhookHeaders(BaseModel):
    event: str = Field(alias='x-github-event')
    hook_id: int = Field(alias='x-github-hook-id')
    delivery: str = Field(alias='x-github-delivery')
    signature: Optional[str] = Field(alias='x-hub-signature')
    signature_256: Optional[str] = Field(alias='x-hub-signature-256')


class WebhookCommonPayload(BaseModel):
    sender: User
    repository: Optional[Repository]
    organization: Optional[Organization]



from typing import Any, Optional, Protocol, Union

from pydantic import BaseModel

from github_webhooks.schemas import WebhookHeaders

PayloadT = type[BaseModel]
HandlerResult = Optional[str]


class HandlerBasic(Protocol):
    async def __call__(self, payload: Any) -> HandlerResult:
        # actually payload will be parsed pydantic model
        ...


class HandlerWithHeaders(Protocol):
    async def __call__(self, payload: Any, *, headers: WebhookHeaders) -> HandlerResult:
        # actually payload will be parsed pydantic model
        ...


class DefaultHandlerBasic(Protocol):
    async def __call__(self, event: str, payload: bytes) -> HandlerResult:
        ...


class DefaultHandlerWithHeaders(Protocol):
    async def __call__(self, event: str, payload: bytes, *, headers: WebhookHeaders) -> HandlerResult:
        ...


Handler = Union[HandlerBasic, HandlerWithHeaders]
DefaultHandler = Union[DefaultHandlerBasic, DefaultHandlerWithHeaders]
AnyHandlerWithHeaders = Union[HandlerWithHeaders, DefaultHandlerWithHeaders]
AnyHandlerWithoutHeaders = Union[HandlerBasic, DefaultHandlerBasic]
